from unittest.mock import MagicMock

from nessai.samplers.importancesampler import ImportanceNestedSampler as INS


def test_update_history(ins, history):
    ins.history = history
    ins.log_likelihood_threshold = -1.23
    ins.logX = -0.5
    ins.gradient = 0.9
    ins.samples_entropy = 1.1
    ins.current_proposal_entropy = 0.8
    ins.live_points_ess = 300

    ins.stopping_criteria = ["ess", "ratio"]
    ins.state = MagicMock()
    ins.state.ess = 500
    ins.state.ratio = 1.0

    INS.update_history(ins)

    assert ins.history["logL_threshold"][-1] == -1.23
    assert ins.history["logX"][-1] == -0.5
    assert ins.history["gradients"][-1] == 0.9
    assert ins.history["samples_entropy"][-1] == 1.1
    assert ins.history["proposal_entropy"][-1] == 0.8
    assert ins.history["live_points_ess"][-1] == 300
    assert ins.history["stopping_criteria"]["ess"][-1] == 500
    assert ins.history["stopping_criteria"]["ratio"][-1] == 1.0



"""Base proposal class that contains common methods."""

import copy
import logging
import os
import re
from abc import abstractmethod
from inspect import signature
from typing import Optional
from warnings import warn

import matplotlib.pyplot as plt
import numpy as np
import numpy.lib.recfunctions as rfn
import torch

from ... import config
from ...flowmodel import FlowModel
from ...livepoint import (
    empty_structured_array,
    get_dtype,
    live_points_to_array,
    numpy_array_to_live_points,
)
from ...plot import nessai_style, plot_1d_comparison, plot_live_points
from ...reparameterisations import (
    CombinedReparameterisation,
    ReparameterisationError,
    get_reparameterisation,
)
from ...utils import (
    save_live_points,
)
from ..rejection import RejectionProposal

logger = logging.getLogger(__name__)


class BaseFlowProposal(RejectionProposal):
    """
    Object that handles training and proposal points

    Parameters
    ----------
    model : :obj:`nessai.model.Model`
        User defined model.
    flow_config : dict, optional
        Configuration for training the normalising flow. If None, uses default
        settings. Defaults to None.
    output : str, optional
        Path to output directory.
    plot : {True, False, 'all', 'min'}, optional
        Controls the plotting level: ``True`` - all plots; ``False`` - no
        plots; ``'all'`` -  all plots and ``'min'`` -  1d plots and loss.
    poolsize : int, optional
        Size of the proposal pool. Defaults to 10000.
    update_poolsize : bool, optional
        If True the poolsize is updated using the current acceptance of the
        nested sampler.
    max_poolsize_scale : int, optional
        Maximum scale for increasing the poolsize. E.g. if this value is 10
        and the poolsize is 1000 the maximum number of points in the pool
        is 10,000.
    check_acceptance : bool, optional
        If True the acceptance is computed after populating the pool. This
        includes computing the likelihood for every point. Default False.
    reparameterisations : Union[dict, str], optional
        Dictionary for configure more flexible reparameterisations. This
        ignores any of the other settings related to rescaling. For more
        details see the documentation.
    fallback_reparameterisation : None or str
        Name of the reparameterisation to be used for parameters that have not
        been specified in the reparameterisations dictionary. If None, the
        :py:class:`~nessai.reparameterisations.NullReparameterisation` is used.
        Reparameterisation should support multiple parameters.
    use_default_reparameterisations : bool, optional
        If True then reparameterisations will be used even if
        ``reparameterisations`` is None. The exact reparameterisations used
        will depend on
        :py:func:`~nessai.proposal.flowproposal.FlowProposal.add_default_reparameterisations`
        which may be overloaded by child classes. If not specified then the
        value of the attribute
        :py:attr:`~nessai.proposal.flowproposal.FlowProposal.use_default_reparameterisations`
        is used.
    reverse_reparameterisations : bool
        Passed to :code:`reverse_order` in
        :py:obj:`~nessai.reparameterisations.combined.CombinedReparameterisation`.
        Reverses the order of the reparameterisations.
    map_to_unit_hypercube : bool
        If true, samples will be mapped to the unit hypercube before any
        reparameterisations are applied.
    """

    use_default_reparameterisations = False
    """
    Indicates whether reparameterisations will be used be default in this
    class. Child classes can change this value a force the default
    behaviour to change without changing the keyword arguments.
    """
    _FlowModelClass = FlowModel
    alt_dist = None

    def __init__(
        self,
        model,
        rng: Optional[np.random.Generator] = None,
        flow_config=None,
        training_config=None,
        output=None,
        poolsize=None,
        plot="min",
        check_acceptance=False,
        max_poolsize_scale=10,
        update_poolsize=True,
        accumulate_weights=False,
        save_training_data=False,
        reparameterisations=None,
        fallback_reparameterisation="zscore",
        use_default_reparameterisations=None,
        reverse_reparameterisations=False,
        map_to_unit_hypercube=False,
    ):
        super().__init__(model, rng=rng)

        self._x_dtype = None
        self._x_prime_dtype = None
        self._draw_func = None
        self._prior_bounds = None

        self.flow = None
        self._flow_config = None
        self._training_config = None
        self.populated = False
        self.populating = False  # Flag used for resuming during population
        self.indices = []
        self.training_count = 0
        self.populated_count = 0
        self.parameters = None
        self.training_data = None
        self.save_training_data = save_training_data
        self.x = None
        self.samples = None
        self.prime_parameters = None
        self.acceptance = []
        self._reparameterisation = None
        self.rescaling_set = False
        self.use_x_prime_prior = False
        self.should_update_reparameterisations = False
        self.map_to_unit_hypercube = map_to_unit_hypercube
        self.accumulate_weights = accumulate_weights

        self.reparameterisations = reparameterisations
        if use_default_reparameterisations is not None:
            self.use_default_reparameterisations = (
                use_default_reparameterisations
            )
        self.fallback_reparameterisation = fallback_reparameterisation
        self.reverse_reparameterisations = reverse_reparameterisations

        self.output = output if output is not None else os.getcwd()

        self.check_acceptance = check_acceptance
        self.flow_config = flow_config
        self.training_config = training_config

        self.configure_poolsize(
            poolsize=poolsize,
            max_poolsize_scale=max_poolsize_scale,
            update_poolsize=update_poolsize,
        )
        self.configure_plotting(plot)

    @property
    def poolsize(self):
        """
        Return the poolsize based of the base value and the current
        value of the scaling
        """
        return int(self._poolsize_scale * self._poolsize)

    @property
    def flow_config(self):
        """Return the configuration for the flow"""
        return self._flow_config

    @flow_config.setter
    def flow_config(self, config):
        """Set configuration.

        Does not update with the defaults because the number of inputs
        has not been determined yet.
        """
        if config is None:
            config = {}
        self._flow_config = config

    @property
    def training_config(self):
        """Return the configuration for the flow"""
        return self._training_config

    @training_config.setter
    def training_config(self, config):
        """Set training configuration."""
        if config is None:
            config = {}
        self._training_config = config

    @property
    def dims(self):
        """Return the number of dimensions"""
        return len(self.parameters)

    @property
    def rescaled_dims(self):
        """Return the number of rescaled dimensions"""
        return len(self.prime_parameters)

    @property
    def x_dtype(self):
        """Return the dtype for the x space"""
        if self._x_dtype is None:
            self._x_dtype = get_dtype(
                self.parameters, config.livepoints.default_float_dtype
            )
        return self._x_dtype

    @property
    def x_prime_dtype(self):
        """Return the dtype for the x prime space"""
        if self._x_prime_dtype is None:
            self._x_prime_dtype = get_dtype(
                self.prime_parameters, config.livepoints.default_float_dtype
            )
        return self._x_prime_dtype

    @property
    def population_dtype(self):
        """
        dtype used for populating the proposal, depends on if the prior
        is defined in the x space or x-prime space
        """
        if self.use_x_prime_prior:
            return self.x_prime_dtype
        else:
            return self.x_dtype

    @property
    def prior_bounds(self):
        """The priors bounds used when computing the priors.

        If :code:`map_to_unit_hypercube` is true, these will be [0, 1]
        """
        if self._prior_bounds is None:
            if self.map_to_unit_hypercube:
                logger.debug("Setting prior bounds to the unit-hypercube")
                self._prior_bounds = {
                    n: np.array([0.0, 1.0]) for n in self.model.names
                }
            else:
                logger.debug("Setting prior bounds to the model prior bounds")
                self._prior_bounds = self.model.bounds
        return self._prior_bounds

    def configure_poolsize(
        self,
        poolsize,
        update_poolsize,
        max_poolsize_scale,
    ):
        """
        Configure settings related to the pool size
        """
        if poolsize is None:
            raise RuntimeError("Must specify `poolsize`")

        self._poolsize = poolsize
        self._poolsize_scale = 1.0
        self.update_poolsize = update_poolsize
        self.max_poolsize_scale = max_poolsize_scale
        self.ns_acceptance = 1.0

    def configure_plotting(self, plot):
        """Configure plotting.

        Plotting is split into training and pool. Training refers to plots
        produced during training and pool refers to plots produces during
        the population stage.

        Parameters
        ----------
        plot : {True, False, 'all', 'train', 'pool', 'min', 'minimal'}
            Level of plotting. `all`, `train` and `pool` enable corner style
            plots. All other values that evaluate to True enable 1d histogram
            plots. False disables all plotting.
        """
        if plot:
            if isinstance(plot, str):
                if plot == "all":
                    self._plot_pool = "all"
                    self._plot_training = "all"
                elif plot == "train":
                    self._plot_pool = False
                    self._plot_training = "all"
                elif plot == "pool":
                    self._plot_pool = "all"
                    self._plot_training = False
                elif plot == "minimal" or plot == "min":
                    self._plot_pool = True
                    self._plot_training = True
                else:
                    logger.warning(
                        f"Unknown plot argument: {plot}, setting all false"
                    )
                    self._plot_pool = False
                    self._plot_training = False
            else:
                self._plot_pool = True
                self._plot_training = True

        else:
            self._plot_pool = False
            self._plot_training = False

    def update_flow_config(self):
        """Update the flow configuration dictionary."""
        self.flow_config["n_inputs"] = self.rescaled_dims

    def initialise(self, resumed: bool = False) -> None:
        """
        Initialise the proposal class.

        This includes:
            * Setting up the rescaling
            * Verifying the rescaling is invertible
            * Initialising the FlowModel

        Parameters
        ----------
        resumed : bool
            Indicates if the proposal is being initialised after being resumed
            or not. When true, the reparameterisations will not be
            reinitialised.
        """
        if not os.path.exists(self.output):
            os.makedirs(self.output, exist_ok=True)

        # Initialise if not resuming or resuming but initialised is False
        if not resumed or not self.initialised:
            self.set_rescaling()
            self.verify_rescaling()

        self.update_flow_config()
        self.flow = self._FlowModelClass(
            flow_config=self.flow_config,
            training_config=self.training_config,
            output=self.output,
            rng=self.rng,
        )
        self.flow.initialise()
        self.populated = False
        self.initialised = True

    def update_poolsize_scale(self, acceptance):
        """
        Update poolsize given the current acceptance.

        Parameters
        ----------
        acceptance : float
            Current acceptance.
        """
        logger.debug(f"Updating poolsize with acceptance: {acceptance:.3f}")
        if not acceptance:
            logger.warning("Acceptance is zero, using maximum scale")
            self._poolsize_scale = self.max_poolsize_scale
        else:
            self._poolsize_scale = 1.0 / acceptance
            if self._poolsize_scale > self.max_poolsize_scale:
                logger.debug("Poolsize scaling is greater than maximum value")
                self._poolsize_scale = self.max_poolsize_scale
            if self._poolsize_scale < 1.0:
                self._poolsize_scale = 1.0

    def add_default_reparameterisations(self):
        """Add any reparameterisations which are assumed by default"""
        logger.debug("No default reparameterisations")

    def get_reparameterisation(self, name):
        """Get the reparameterisation from the name"""
        return get_reparameterisation(name)

    def configure_reparameterisations(self, reparameterisations):
        """Configure the reparameterisations.

        Parameters
        ----------
        reparameterisations : Union[dict, str, None]
            Dictionary of reparameterisations. If None, then the defaults
            from :py:func`get_default_reparameterisations` are used.
        """
        if reparameterisations is None:
            logger.info(
                "No reparameterisations provided, using default "
                f"reparameterisations included in {self.__class__.__name__}"
            )
            _reparameterisations = {}
        else:
            _reparameterisations = copy.deepcopy(reparameterisations)
        logger.info(f"Adding reparameterisations from: {_reparameterisations}")
        self._reparameterisation = CombinedReparameterisation(
            reverse_order=self.reverse_reparameterisations
        )
        if isinstance(_reparameterisations, str):
            _reparameterisations = {
                _reparameterisations: {"parameters": self.model.names}
            }
        elif not isinstance(_reparameterisations, dict):
            raise TypeError(
                "Reparameterisations must be a dictionary, string or None, "
                f"received {type(_reparameterisations).__name__}"
            )

        for k, cfg in _reparameterisations.items():
            if k in self.model.names:
                logger.debug(
                    f"Found parameter {k} in model, assuming it is a parameter"
                )
                if isinstance(cfg, str) or cfg is None:
                    rc, default_config = self.get_reparameterisation(cfg)
                    default_config["parameters"] = k
                elif isinstance(cfg, dict):
                    if cfg.get("reparameterisation", None) is None:
                        raise RuntimeError(
                            f"No reparameterisation found for {k}. "
                            "Check inputs (and their spelling :)). "
                            f"Current keys: {list(cfg.keys())}"
                        )
                    rc, default_config = self.get_reparameterisation(
                        cfg["reparameterisation"]
                    )
                    cfg.pop("reparameterisation")

                    if cfg.get("parameters", False):
                        cfg["parameters"] += [k]
                    else:
                        default_config["parameters"] = k

                    default_config.update(cfg)
                else:
                    raise TypeError(
                        f"Unknown config type for: {k}. Expected str or dict, "
                        f"received instance of {type(cfg)}."
                    )
            else:
                logger.debug(f"Assuming {k} is a reparameterisation")
                try:
                    rc, default_config = self.get_reparameterisation(k)
                    if isinstance(cfg, list):
                        logger.debug("Assuming list of patterns")
                        cfg = {"parameters": cfg}
                    default_config.update(cfg)
                    parameters = default_config.get("parameters")

                    if parameters is not None:
                        if not isinstance(parameters, list):
                            if isinstance(parameters, str):
                                patterns = [parameters]
                            else:
                                patterns = list(parameters)
                        else:
                            patterns = parameters.copy()
                        matches = []
                        for pattern in patterns:
                            r = re.compile(pattern)
                            matches += list(
                                filter(r.fullmatch, self.model.names)
                            )
                        default_config["parameters"] = matches
                    else:
                        logger.warning(
                            "Reparameterisation might be missing parameters!"
                        )

                except ValueError:
                    raise RuntimeError(
                        f"{k} is not a parameter in the model or a known "
                        "reparameterisation"
                    )

            if not default_config.get("parameters", False):
                raise RuntimeError(
                    "No parameters key in the config! "
                    "Check reparameterisations, setting logging"
                    " level to DEBUG can be helpful"
                )

            if (
                "boundary_inversion" in default_config
                and default_config["boundary_inversion"]
            ):
                self.boundary_inversion = True

            if isinstance(default_config["parameters"], list):
                prior_bounds = {
                    p: self.prior_bounds[p]
                    for p in default_config["parameters"]
                }
            else:
                prior_bounds = {
                    default_config["parameters"]: self.prior_bounds[
                        default_config["parameters"]
                    ]
                }
            sig = signature(rc.__init__)
            if "rng" in sig.parameters:
                default_config["rng"] = self.rng

            logger.info(f"Adding {rc.__name__} with config: {default_config}")
            r = rc(prior_bounds=prior_bounds, **default_config)
            self._reparameterisation.add_reparameterisations(r)

        if self.use_default_reparameterisations:
            self.add_default_reparameterisations()

        other_params = [
            n
            for n in self.model.names
            if n not in self._reparameterisation.parameters
        ]
        if other_params:
            logger.debug("Getting fallback reparameterisation")
            FallbackClass, fallback_kwargs = self.get_reparameterisation(
                self.fallback_reparameterisation
            )
            fallback_kwargs["prior_bounds"] = {
                p: self.prior_bounds[p] for p in other_params
            }
            logger.info(
                f"Assuming fallback reparameterisation "
                f"({FallbackClass.__name__}) for {other_params} with kwargs: "
                f"{fallback_kwargs}."
            )
            r = FallbackClass(parameters=other_params, **fallback_kwargs)
            self._reparameterisation.add_reparameterisations(r)

        if any(r._update for r in self._reparameterisation.values()):
            self.should_update_reparameterisations = True
        else:
            self.should_update_reparameterisations = False

        if self._reparameterisation.has_prime_prior:
            warn(
                (
                    "Support for x-prime priors is deprecated and will be "
                    "removed in a future release. "
                ),
                FutureWarning,
            )
            self.use_x_prime_prior = True
            self.x_prime_log_prior = self._reparameterisation.x_prime_log_prior
            logger.debug("Using x prime prior")
            if self.map_to_unit_hypercube:
                raise RuntimeError(
                    "x prime prior does not support map to unit hypercube"
                )
        else:
            logger.debug("Prime prior is disabled")
            if self._reparameterisation.requires_prime_prior:
                raise RuntimeError(
                    "One or more reparameterisations require use of the x "
                    "prime prior but it cannot be enabled with the current "
                    "settings. Note that the x prime prior is deprecated and "
                    "will be removed in a future release."
                )

        self._reparameterisation.check_order()

        self.parameters = self._reparameterisation.parameters
        self.prime_parameters = self._reparameterisation.prime_parameters

    @property
    def names(self):
        warn(
            "`names` is deprecated, use `parameters` instead",
            FutureWarning,
        )
        return self.parameters

    @property
    def rescaled_names(self):
        warn(
            ("`rescaled_names` is deprecated, use `prime_parameters` instead"),
            FutureWarning,
        )
        return self.prime_parameters

    @property
    def update_bounds(self):
        warn(
            (
                "`update_bounds` is deprecated, use "
                "`should_update_reparameterisations` instead."
            ),
            FutureWarning,
        )
        return self.should_update_reparameterisations

    def set_rescaling(self):
        """
        Set function and parameter names for rescaling
        """
        if self.model.reparameterisations is not None:
            self.reparameterisations = self.model.reparameterisations

        self.configure_reparameterisations(self.reparameterisations)

        logger.info(f"x space parameters: {self.parameters}")
        logger.info(f"x prime space parameters: {self.prime_parameters}")
        self.rescaling_set = True

    def verify_rescaling(self):
        """
        Verify the rescaling functions are invertible
        """
        if not self.rescaling_set:
            raise RuntimeError(
                "Rescaling must be set before it can be verified"
            )
        if not self._reparameterisation.one_to_one:
            logger.warning(
                "Could not check if reparameterisation is invertible"
            )
            return
        logger.info("Verifying rescaling functions")
        x = self.model.new_point(N=1000)
        for inversion in ["lower", "upper", False, None]:
            self.check_state(x)
            logger.debug(f"Testing: {inversion}")
            x_prime, log_J = self.rescale(x, test=inversion)
            x_out, log_J_inv = self.inverse_rescale(x_prime)

            n = x.size
            ratio = x_out.size // x.size
            logger.debug(f"Ratio of output to input: {ratio}")
            for f in x.dtype.names:
                target = x[f]
                for count in range(ratio):
                    start = count * n
                    end = (count + 1) * n
                    block = x_out[f][start:end]
                    if f in config.livepoints.non_sampling_parameters:
                        if not np.allclose(block, target, equal_nan=True):
                            raise RuntimeError(
                                f"Non-sampling parameter {f} changed in "
                                f" the rescaling (block {count})."
                            )
                    elif not np.allclose(block, target, equal_nan=False):
                        msg = f"Rescaling is not invertible for {f}! "
                        if ratio > 1:
                            msg += f"(block {count}) "
                        log_dynamic_range = (
                            np.log10(block).max() - np.log10(block).min()
                        )
                        if log_dynamic_range > 10:
                            msg += (
                                "This may be due to the large dynamic range "
                                f"(log10 dynamic range={log_dynamic_range:.1f}, "
                                f"min={block.min()}, "
                                f"max={block.max()}). \n"
                                "See https://nessai.readthedocs.io/en/latest/faqs.html#large-dynamic-ranges "
                                "for a possible solution."
                            )
                        raise ReparameterisationError(msg)
                    else:
                        logger.debug(f"Block {count} is equal to the input")
            if not np.allclose(log_J, -log_J_inv):
                raise RuntimeError("Rescaling Jacobian is not invertible")

        logger.info("Rescaling functions are invertible")
        self._reparameterisation.reset()

    def rescale(self, x, **kwargs):
        """
        Rescale from the physical space to the primed physical space

        Parameters
        ----------
        x: array_like
            Array of live points to rescale
        compute_radius: bool (False)
            Used to indicate when rescaling is being used for computing the
            radius for population. This is important for rescaling that uses
            inversions.

        Returns
        -------
        array
            Array of rescaled values
        array
            Array of log det|J|
        """
        x_prime = empty_structured_array(x.size, dtype=self.x_prime_dtype)
        log_J = np.zeros(x_prime.size)

        if x.size == 1:
            x = np.array([x], dtype=x.dtype)

        if self.map_to_unit_hypercube:
            x = self.model.to_unit_hypercube(x)

        x, x_prime, log_J = self._reparameterisation.reparameterise(
            x, x_prime, log_J, **kwargs
        )

        for p in config.livepoints.non_sampling_parameters:
            x_prime[p] = x[p]
        return x_prime, log_J

    def inverse_rescale(self, x_prime, return_unit_hypercube=False, **kwargs):
        """
        Rescale from the primed physical space to the original physical
        space.

        Parameters
        ----------
        x_prime : array_like
            Array of live points to rescale.

        Returns
        -------
        array
            Array of rescaled values in the data space.
        array
            Array of log-Jacobian determinants.
        """
        x = empty_structured_array(x_prime.size, dtype=self.x_dtype)
        log_J = np.zeros(x.size)
        x, x_prime, log_J = self._reparameterisation.inverse_reparameterise(
            x, x_prime, log_J, **kwargs
        )

        for p in config.livepoints.non_sampling_parameters:
            x[p] = x_prime[p]

        if self.map_to_unit_hypercube and not return_unit_hypercube:
            x = self.model.from_unit_hypercube(x)

        return x, log_J

    def check_state(self, x):
        """Update the state of the proposal given some training data.

        Includes updating the reparameterisations.

        Parameters
        ----------
        x: array_like
            Array of training live points which can be used to set parameters
        """
        if self.map_to_unit_hypercube:
            x = self.model.to_unit_hypercube(x)
        self._reparameterisation.update(x)

    @nessai_style()
    def _plot_training_data(self, output):
        """Plot the training data and compare to the results"""
        z_training_data, _ = self.forward_pass(
            self.training_data, rescale=True
        )
        z_gen = self.flow.sample_latent_distribution(self.training_data.size)

        fig = plt.figure()
        plt.hist(np.sqrt(np.sum(z_training_data**2, axis=1)), "auto")
        plt.xlabel("Radius")
        fig.savefig(os.path.join(output, "radial_dist.png"))
        plt.close(fig)

        plot_1d_comparison(
            z_training_data,
            z_gen,
            labels=["z_live_points", "z_generated"],
            convert_to_live_points=True,
            filename=os.path.join(output, "z_comparison.png"),
        )

        x_prime_gen, log_prob = self.backward_pass(z_gen, rescale=False)
        x_prime_gen["logL"] = log_prob
        x_gen, log_J = self.inverse_rescale(x_prime_gen)
        x_gen, log_J, x_prime_gen = self.check_prior_bounds(
            x_gen, log_J, x_prime_gen
        )
        x_gen["logL"] += log_J

        plot_1d_comparison(
            self.training_data,
            x_gen,
            parameters=self.model.names,
            labels=["live points", "generated"],
            filename=os.path.join(output, "x_comparison.png"),
        )

        if self._plot_training == "all":
            plot_live_points(
                self.training_data,
                c="logL",
                filename=os.path.join(output, "x_samples.png"),
            )

            plot_live_points(
                x_gen,
                c="logL",
                filename=os.path.join(output, "x_generated.png"),
            )

            plot_live_points(
                self.training_data_prime,
                c="logL",
                filename=os.path.join(output, "x_prime_samples.png"),
            )
            plot_live_points(
                x_prime_gen,
                c="logL",
                filename=os.path.join(output, "x_prime_generated.png"),
            )

        plot_1d_comparison(
            self.training_data_prime,
            x_prime_gen,
            parameters=self.prime_parameters,
            labels=["live points", "generated"],
            filename=os.path.join(output, "x_prime_comparison.png"),
        )

    def train(self, x, plot=True):
        """
        Train the normalising flow given some of the live points.

        Parameters
        ----------
        x : structured_array
            Array of live points
        plot : {True, False, 'all'}
            Enable or disable plots for during training. By default the plots
            are only one-dimensional histograms, `'all'` includes corner
            plots with samples, these are often a few MB in size so
            proceed with caution!
        """
        if not self.initialised:
            raise RuntimeError(f"{self.__name__} is not initialised.")

        if (plot and self._plot_training) or self.save_training_data:
            block_output = os.path.join(
                self.output, "training", f"block_{self.training_count}", ""
            )
        else:
            block_output = self.output

        if not os.path.exists(block_output):
            os.makedirs(block_output, exist_ok=True)

        if self.save_training_data:
            save_live_points(
                x, os.path.join(block_output, "training_data.json")
            )

        self.training_data = x.copy()
        self.check_state(self.training_data)

        x_prime, _ = self.rescale(x)

        self.training_data_prime = x_prime.copy()

        # Convert to numpy array for training and remove likelihoods and priors
        # Since the names of parameters may have changes, pull names from flows
        x_prime_array = live_points_to_array(
            x_prime, self.prime_parameters, copy=True
        )

        self.flow.train(
            x_prime_array,
            output=block_output,
            plot=self._plot_training and plot,
        )

        if self._plot_training and plot:
            self._plot_training_data(block_output)

        self.populated = False
        self.training_count += 1

    def reset_model_weights(self, **kwargs):
        """
        Reset the flow weights.

        Parameters
        ----------
        kwargs :
            Keyword arguments passed to
            :meth:`nessai.flowmodel.FlowModel.reset_model`.
        """
        self.flow.reset_model(**kwargs)

    def check_prior_bounds(self, x, *args):
        """
        Return only values that are within the prior bounds

        Parameters
        ----------
        x: array_like
            Array of live points which will compared against prior bounds
        *args:
            Additional arrays which correspond to the array of live points.
            Only those corresponding to points within the prior bounds
            are returned

        Returns
        -------
        out: tuple of arrays
            Array containing the subset of the original arrays which fall
            within the prior bounds
        """
        flags = self.model.in_bounds(x)
        return (a[flags] for a in (x,) + args)

    def forward_pass(self, x, rescale=True, **kwargs):
        """
        Pass a vector of points through the model

        Parameters
        ----------
        x : array_like
            Live points to map to the latent space
        rescale : bool, optional (True)
            Apply rescaling function
        compute_radius : bool, optional (True)
            Flag parsed to rescaling for rescaling specific to radius
            computation

        Returns
        -------
        x : array_like
            Samples in the latent space
        log_prob : array_like
            Log probabilities corresponding to each sample (including the
            jacobian)
        """
        log_J = 0
        if rescale:
            x, log_J_rescale = self.rescale(x, **kwargs)
            log_J += log_J_rescale

        x = live_points_to_array(x, names=self.prime_parameters, copy=True)

        if x.ndim == 1:
            x = x[np.newaxis, :]
        z, log_prob = self.flow.forward_and_log_prob(x)

        return z, log_prob + log_J

    def backward_pass(self, z, rescale=True, **kwargs):
        """
        A backwards pass from the model (latent -> real)

        Parameters
        ----------
        z : array_like
            Structured array of points in the latent space
        rescale : bool, optional (True)
            Apply inverse rescaling function

        Returns
        -------
        x : array_like
            Samples in the data space
        log_j : array_like
            Log Jacobian determinant
        """
        # Compute the log probability
        x, log_j = self.flow.inverse(z)

        x = numpy_array_to_live_points(
            x.astype(config.livepoints.default_float_dtype),
            self.prime_parameters,
        )
        # Apply rescaling in rescale=True
        if rescale:
            x, log_j_rescale = self.inverse_rescale(x, **kwargs)
            # Include Jacobian for the rescaling
            log_j += log_j_rescale
        return x, log_j

    def log_prior(self, x):
        """
        Compute the prior probability using the user-defined model

        Parameters
        ----------
        x : structured_array
            Array of samples

        Returns
        -------
        array_like
            Array of log prior probabilities
        """
        if self._reparameterisation:
            return self.model.batch_evaluate_log_prior(
                x
            ) + self._reparameterisation.log_prior(x)
        else:
            return self.model.batch_evaluate_log_prior(x)

    def x_prime_log_prior(self, x):
        """
        Compute the prior in the prime space

        Parameters
        ----------
        x : array
            Samples in the X-prime space.
        """
        raise RuntimeError("Prime prior is not implemented")

    def unit_hypercube_log_prior(self, x):
        """
        Compute the prior in the unit hypercube space.

        Parameters
        ----------
        x : array
            Samples in the unit hypercube.
        """
        if self._reparameterisation:
            return self.model.batch_evaluate_log_prior_unit_hypercube(
                x
            ) + self._reparameterisation.log_prior(x)
        else:
            return self.model.batch_evaluate_log_prior_unit_hypercube(x)

    def compute_weights(self, x, log_q, return_log_prior=False):
        """
        Compute weights for the samples.

        Does NOT normalise the weights

        Parameters
        ----------
        x :  structured_arrays
            Array of points
        log_q : array_like
            Array of log proposal probabilities.
        return_log_prior: bool
            If true, the log-prior probability is also returned.

        Returns
        -------
        array_like
            Log-weights for rejection sampling.
        """
        if self.use_x_prime_prior:
            log_p = self.x_prime_log_prior(x)
        elif self.map_to_unit_hypercube:
            log_p = self.unit_hypercube_log_prior(x)
        else:
            log_p = self.log_prior(x)

        log_w = log_p - log_q
        if return_log_prior:
            return log_w, log_p
        else:
            return log_w

    def convert_to_samples(self, x, plot=True):
        """
        Convert the array to samples ready to be used.

        This removes are auxiliary parameters, (e.g. auxiliary radial
        parameters) and ensures the prior is computed. These samples can
        be directly used in the nested sampler.

        Parameters
        ----------
        x : array_like
            Array of samples
        plot : bool, optional
            If true a 1d histogram for each parameter of the pool is plotted.
            This includes a comparison the live points used to train the
            current realisation of the flow.

        Returns
        -------
        array
            Structured array of samples.
        """
        if self.use_x_prime_prior:
            if self._plot_pool and plot:
                plot_1d_comparison(
                    self.training_data_prime,
                    x,
                    labels=["live points", "pool"],
                    filename=os.path.join(
                        self.output, f"pool_prime_{self.populated_count}.png"
                    ),
                )

            x, _ = self.inverse_rescale(x)
        elif self.map_to_unit_hypercube:
            x = self.model.from_unit_hypercube(x)
        x = rfn.repack_fields(
            x[self.model.names + config.livepoints.non_sampling_parameters]
        )
        x["logP"] = self.model.batch_evaluate_log_prior(x)
        return x

    @abstractmethod
    def populate(self, worst_point, n_samples=10000):
        raise NotImplementedError

    def compute_acceptance(self, logL):
        """
        Compute how many of the current pool have log-likelihoods greater
        than the specified log-likelihood using the current value in the
        `logL` field.

        Parameters
        ----------
        float : logL
            Log-likelihood to use as the lower bound

        Returns
        -------
        float :
            Acceptance defined on [0, 1]
        """
        return (self.samples["logL"] > logL).sum() / self.samples.size

    def draw(self, worst_point):
        """
        Draw a replacement point. The new point is independent of the worst
        point. The worst point is only used during population.

        Parameters
        ----------
        worst_point : structured_array
            The current worst point used to compute the radius of the contour
            in the latent space.

        Returns
        -------
        structured_array
            New live point
        """
        if not self.populated:
            self.populating = True
            if self.update_poolsize:
                self.update_poolsize_scale(self.ns_acceptance)
            while not self.populated:
                self.populate(worst_point, n_samples=self.poolsize)
            self.populating = False
        # new sample is drawn randomly from proposed points
        # popping from right end is faster
        index = self.indices.pop()
        new_sample = self.samples[index]
        if not self.indices:
            self.populated = False
            logger.debug("Proposal pool is empty")
        # make live point and return
        return new_sample

    @nessai_style()
    def plot_pool(self, x):
        """
        Plot the pool of points.

        Parameters
        ----------
        x : array_like
            Corresponding samples to plot in the physical space.
        """
        if self._plot_pool == "all":
            plot_live_points(
                x,
                c="logL",
                filename=os.path.join(
                    self.output, f"pool_{self.populated_count}.png"
                ),
            )
        else:
            plot_1d_comparison(
                self.training_data,
                x,
                labels=["live points", "pool"],
                filename=os.path.join(
                    self.output, f"pool_{self.populated_count}.png"
                ),
            )

            z, log_q = self.forward_pass(x, compute_radius=False)
            z_tensor = (
                torch.from_numpy(z)
                .type(torch.get_default_dtype())
                .to(self.flow.device)
            )
            with torch.inference_mode():
                if self.alt_dist is not None:
                    log_p = self.alt_dist.log_prob(z_tensor).cpu().numpy()
                else:
                    log_p = (
                        self.flow.model.base_distribution_log_prob(z_tensor)
                        .cpu()
                        .numpy()
                    )

            fig, axs = plt.subplots(3, 1, figsize=(3, 9))
            axs = axs.ravel()
            axs[0].hist(log_q, 20, histtype="step", label="log q")
            axs[1].hist(log_q - log_p, 20, histtype="step", label="log J")
            axs[2].hist(
                np.sqrt(np.sum(z**2, axis=1)),
                20,
                histtype="step",
                label="Latent radius",
            )
            axs[0].set_xlabel("Log q")
            axs[1].set_xlabel("Log |J|")
            axs[2].set_xlabel("r")
            plt.tight_layout()
            fig.savefig(
                os.path.join(
                    self.output, f"pool_{self.populated_count}_log_q.png"
                )
            )
            plt.close(fig)

    def resume(self, model, flow_config, weights_file=None):
        """
        Resume the proposal.

        The model and config are not stored so these must be provided.

        Parameters
        ----------
        model : :obj:`nessai.model.Model`
            User-defined model used.
        flow_config : dict
            Configuration dictionary for the flow.
        weights_files : str, optional
            Weights file to try and load. If not provided the proposal
            tries to load the last weights file.
        """
        super().resume(model)
        self.flow_config = flow_config

        if self.mask is not None:
            if isinstance(self.mask, list):
                m = np.array(self.mask)
            self.flow_config["mask"] = m

        self.initialise(resumed=True)

        if weights_file is None:
            weights_file = self.weights_file

        # Flow might have exited before any weights were saved.
        if weights_file is not None:
            if os.path.exists(weights_file):
                self.flow.reload_weights(weights_file)
        else:
            logger.warning("Could not reload weights for flow")

    def reset(self):
        """Reset the proposal"""
        self.indices = []
        self.samples = None
        self.x = None
        self.populated = False
        self.populated_count = 0
        self.population_acceptance = None
        self._poolsize_scale = 1.0
        self._checked_population = True
        self.acceptance = []
        self._reparameterisation.reset()

    def __getstate__(self):
        state = self.__dict__.copy()
        state["initialised"] = False
        state["weights_file"] = getattr(
            state.get("flow"), "weights_file", None
        )

        # Mask may be generate via permutation, so must be saved
        if "mask" in getattr(state.get("flow"), "flow_config", {}):
            state["mask"] = state["flow"].flow_config["mask"]
        else:
            state["mask"] = None
        if state["populated"] and state["indices"]:
            state["resume_populated"] = True
        else:
            state["resume_populated"] = False

        # user provides model and config for resume
        # flow can be reconstructed from resume
        del state["model"]
        del state["_flow_config"]
        del state["flow"]

        return state



#!/usr/bin/env python

# Example of using nessai with `reparameterisations` dictionary. This example
# uses the same model as the half_gaussian example.

import numpy as np
from scipy.stats import norm

from nessai.flowsampler import FlowSampler
from nessai.model import Model
from nessai.utils import configure_logger

output = "./outdir/reparameterisations_example/"
logger = configure_logger(output=output)


class HalfGaussian(Model):
    """Two-dimensional Gaussian with a bound at y=0."""

    def __init__(self):
        self.names = ["x", "y"]
        self.bounds = {"x": [-10, 10], "y": [0, 10]}

    def log_prior(self, x):
        """Log-prior"""
        log_p = np.log(self.in_bounds(x), dtype="float")
        for bounds in self.bounds.values():
            log_p -= np.log(bounds[1] - bounds[0])
        return log_p

    def log_likelihood(self, x):
        """Log-likelihood"""
        log_l = np.zeros(x.size)
        for n in self.names:
            log_l += norm.logpdf(x[n])
        return log_l


# In this example we use the reparameterisation options to specify how each
# parameter should be rescaled.
# In this case we'll tell nessai to use an inversion for y, but only at the
# lower bound.
fs = FlowSampler(
    HalfGaussian(),
    output=output,
    resume=False,
    seed=1234,
    reparameterisations={
        "x": "default",
        "y": {
            "reparameterisation": "inversion",
            "detect_edges_kwargs": {"allowed_bounds": ["lower"]},
        },
    },
)

fs.run()



#!/usr/bin/env python

# Example of using the experimental MCMCFlowProposal

import numpy as np

from nessai.flowsampler import FlowSampler
from nessai.model import Model
from nessai.utils import configure_logger

output = "./outdir/mcmc_example_cvm_8d_flow_act/"
logger = configure_logger(output=output)


class RosenbrockModel(Model):
    """Rosenbrock likelihood defined in n dimensions on [-5, 5]^n."""

    def __init__(self, dims):
        self.names = [f"x_{d}" for d in range(dims)]
        self.bounds = {n: [-5.0, 5.0] for n in self.names}

    def log_prior(self, x):
        # The MCMC sampler can return samples outside the prior so we suppress
        # the warnings
        with np.errstate(divide="ignore"):
            log_p = np.log(self.in_bounds(x), dtype="float")
        for bounds in self.bounds.values():
            log_p -= np.log(bounds[1] - bounds[0])
        return log_p

    def log_likelihood(self, x):
        # Unstructured view returns a view of the inputs as a "normal" numpy
        # array without any fields. It will included the parameters listed in
        # names and in the same order. This is faster then converting from
        # live points to an array.
        x = self.unstructured_view(x)
        return -np.sum(
            100.0 * (x[..., 1:] - x[..., :-1] ** 2.0) ** 2.0
            + (1.0 - x[..., :-1]) ** 2.0,
            axis=-1,
        )


# The MCMCFlowProposal class shares most of the configuration options with
# FlowProposal but also has some specific options
# Here we also set
#  - n_accept : the number of accepted jumps required to stop
#  - reset_flow : this is recommended with the MCMC proposal
fs = FlowSampler(
    RosenbrockModel(8),
    output=output,
    resume=False,
    seed=1234,
    flow_proposal_class="mcmcflowproposal",
    plot_history=True,
    plot_chain=False,
    step_type="diff",
    enforce_likelihood_threshold=True,
    # constant_volume_fraction=0.98,
    n_steps=5000,
    reset_flow=16,
    flow_config=dict(
        n_neurons=32,
        n_blocks=8,
        n_layers=2,
        batch_norm_between_layers=True,
        linear_transform=None,
        net="mlp",
    ),
)

# And go!
fs.run()



# -*- coding: utf-8 -*-
"""Utilities for proposal classes"""

import logging
from copy import deepcopy
from inspect import getmro, signature
from typing import Callable, Union
from warnings import warn

from ..utils.settings import _get_all_kwargs

logger = logging.getLogger(__name__)


def check_proposal_kwargs(ProposalClass, kwargs, strict=False):
    """Check the keyword arguments are correct for a proposal class.

    Removes any keyword arguments that correspond to another class. Raises an
    error if the keyword arguments are unknown. For example if the class is
    FlowProposal, it will remove any keyword arguments that correspond to
    AugmentedFlowProposal but not raise an error.

    Parameters
    ----------
    ProposalClass : Type[FlowProposal]
        Class to check the keyword arguments against.
    kwargs : dict
        Dictionary of keyword arguments.
    strict : bool
        Raise an error if any extra keyword arguments are provided even if they
        correspond to other proposals.

    Returns
    -------
    dict
        Dictionary of updated kwargs.
    """
    proposals = list(available_base_flow_proposal_classes().values()) + list(
        available_external_flow_proposal_classes(load=True).values()
    )

    class_keys = set()
    for cls in getmro(ProposalClass):
        class_keys.update(signature(cls).parameters.keys())
        if cls in proposals:
            proposals.remove(cls)

    keys = set(kwargs.keys())
    kwargs_out = deepcopy(kwargs)

    extra_keys = keys - class_keys

    if not extra_keys:
        logger.debug("All keyword arguments match the proposal class")
        return kwargs_out
    elif strict:
        raise RuntimeError(
            f"Keyword arguments contain unknown keys: {extra_keys}"
        )

    allowed_extra_keys = set()

    for proposal in proposals:
        allowed_extra_keys.update(set(_get_all_kwargs(proposal)))

    invalid_keys = extra_keys - allowed_extra_keys

    if invalid_keys:
        raise RuntimeError(
            f"Unknown kwargs for {ProposalClass.__name__}: {invalid_keys}."
        )
    else:
        logger.warning(
            f"Removing unused keyword arguments ({extra_keys}) from kwargs for"
            f" {ProposalClass.__name__}. These are valid keyword arguments but"
            " correspond to other proposal classes."
        )
        for key in extra_keys:
            kwargs_out.pop(key)
    return kwargs_out


def available_base_flow_proposal_classes():
    from ..experimental.gw.proposal import ClusteringGWFlowProposal
    from ..experimental.proposal.clustering import ClusteringFlowProposal
    from ..experimental.proposal.mcmc import MCMCFlowProposal
    from ..gw.proposal import AugmentedGWFlowProposal, GWFlowProposal
    from .augmented import AugmentedFlowProposal
    from .flowproposal import FlowProposal

    base_proposals = {
        "mcmcflowproposal": MCMCFlowProposal,
        "clusteringgwflowproposal": ClusteringGWFlowProposal,
        "augmentedgwflowproposal": AugmentedGWFlowProposal,
        "gwflowproposal": GWFlowProposal,
        "clusteringflowproposal": ClusteringFlowProposal,
        "augmentedflowproposal": AugmentedFlowProposal,
        "flowproposal": FlowProposal,
    }
    return base_proposals


def available_external_flow_proposal_classes(load: bool = False):
    from ..utils.entry_points import get_entry_points

    external_proposals = get_entry_points("nessai.proposals")
    logger.debug(
        f"Found the following external proposals: {external_proposals.keys()}"
    )
    if load:
        for key in external_proposals:
            external_proposals[key] = external_proposals[key].load()
    return external_proposals


def get_flow_proposal_class(
    proposal_class: Union[str, None, Callable],
) -> Callable:
    """Get the proposal class for the standard nested sampler.

    Can also load proposals from entry points. These take priority of the
    default proposals.

    Parameters
    ----------
    proposal_class : Union[str, Proposal, None]
        The name of the proposal class or the class itself. If not specified,
        defaults to :py:obj:`nessai.proposal.flowpropsoal.FlowProposal`.

    Returns
    -------
    Proposal class
    """
    from .flowproposal import FlowProposal

    base_proposals = available_base_flow_proposal_classes()
    external_proposals = available_external_flow_proposal_classes(load=False)

    if proposal_class is None:
        return FlowProposal
    elif isinstance(proposal_class, str):
        proposal_class = proposal_class.lower()
        if proposal_class in external_proposals:
            logger.info("Using external proposal class")
            return external_proposals[proposal_class].load()
        elif proposal_class in base_proposals:
            return base_proposals[proposal_class]
        else:
            raise ValueError(
                f"Unknown proposal class: {proposal_class}. "
                f"Choose from: {list(base_proposals.keys())} or "
                f"{list(external_proposals.keys())}"
            )
    elif issubclass(proposal_class, FlowProposal):
        return proposal_class
    else:
        raise TypeError(
            "Unknown proposal_class type. Must a str, subclass of "
            f"FlowProposal or None. "
            f"Actual input: {proposal_class} ({type(proposal_class)})."
        )


def get_region_sampler_proposal_class(*args, **kwargs):
    warn(
        (
            "`get_region_sampler_proposal_class` is deprecated in favour of "
            "`get_flow_proposal_class`"
        ),
        FutureWarning,
    )
    return get_flow_proposal_class(*args, **kwargs)



"""Test methods related to computing weights"""

import os
from unittest.mock import MagicMock, Mock, patch

import numpy as np
import pytest

from nessai import config
from nessai.livepoint import numpy_array_to_live_points
from nessai.proposal.flowproposal.base import BaseFlowProposal
from nessai.utils.testing import assert_structured_arrays_equal


@pytest.fixture()
def z():
    return np.random.randn(2, 2)


@pytest.fixture()
def x(z):
    return numpy_array_to_live_points(np.random.randn(*z.shape), ["x", "y"])


@pytest.fixture()
def log_q(x):
    return np.random.randn(x.size)


def test_log_prior_wo_reparameterisation(proposal, x):
    """Test the lop prior method"""
    log_prior = -np.ones(x.size)
    proposal._reparameterisation = None
    proposal.model = MagicMock()
    proposal.model.batch_evaluate_log_prior = MagicMock(return_value=log_prior)

    log_prior_out = BaseFlowProposal.log_prior(proposal, x)

    assert np.array_equal(log_prior, log_prior_out)
    proposal.model.batch_evaluate_log_prior.assert_called_once_with(x)


def test_log_prior_w_reparameterisation(proposal, x):
    """Test the lop prior method with reparameterisations"""
    log_prior = -np.ones(x.size)
    proposal._reparameterisation = MagicMock()
    proposal._reparameterisation.log_prior = MagicMock(return_value=log_prior)
    proposal.model = MagicMock()
    proposal.model.batch_evaluate_log_prior = MagicMock(
        return_value=log_prior.copy()
    )

    log_prior_out = BaseFlowProposal.log_prior(proposal, x)

    assert np.array_equal(log_prior_out, -2 * np.ones(x.size))
    proposal._reparameterisation.log_prior.assert_called_once_with(x)
    proposal.model.batch_evaluate_log_prior.assert_called_once_with(x)


def test_prime_log_prior(proposal):
    """Make sure the prime prior raises an error by default."""
    with pytest.raises(RuntimeError) as excinfo:
        BaseFlowProposal.x_prime_log_prior(proposal, 1.0)
    assert "Prime prior is not implemented" in str(excinfo.value)


def test_unit_hypercube_log_prior_wo_reparameterisation(proposal, x):
    log_prior = -np.ones(x.size)
    proposal._reparameterisation = None
    proposal.model = MagicMock()
    proposal.model.batch_evaluate_log_prior_unit_hypercube = MagicMock(
        return_value=log_prior
    )

    log_prior_out = BaseFlowProposal.unit_hypercube_log_prior(proposal, x)

    assert np.array_equal(log_prior, log_prior_out)
    proposal.model.batch_evaluate_log_prior_unit_hypercube.assert_called_once_with(  # noqa: E501
        x
    )


def test_unit_hypercube_log_prior_w_reparameterisation(proposal, x):
    log_prior = -np.ones(x.size)
    proposal._reparameterisation = MagicMock()
    proposal._reparameterisation.log_prior = MagicMock(return_value=log_prior)
    proposal.model = MagicMock()
    proposal.model.batch_evaluate_log_prior_unit_hypercube = MagicMock(
        return_value=log_prior.copy()
    )

    log_prior_out = BaseFlowProposal.unit_hypercube_log_prior(proposal, x)

    assert np.array_equal(log_prior_out, -2 * np.ones(x.size))
    proposal._reparameterisation.log_prior.assert_called_once_with(x)
    proposal.model.batch_evaluate_log_prior_unit_hypercube.assert_called_once_with(  # noqa: E501
        x
    )


@pytest.mark.parametrize(
    "acceptance, scale", [(0.0, 10.0), (0.5, 2.0), (0.01, 10.0), (2.0, 1.0)]
)
def test_update_poolsize_scale(proposal, acceptance, scale):
    """
    Test the check the poolsize is correct adjusted based on the acceptance.
    """
    proposal.max_poolsize_scale = 10.0
    BaseFlowProposal.update_poolsize_scale(proposal, acceptance)
    assert proposal._poolsize_scale == scale


def test_compute_weights(proposal, x, log_q):
    """Test method for computing rejection sampling weights"""
    proposal.use_x_prime_prior = False
    proposal.log_prior = MagicMock(return_value=-np.ones(x.size))
    log_w = BaseFlowProposal.compute_weights(proposal, x, log_q)

    proposal.log_prior.assert_called_once_with(x)
    out = -1 - log_q
    np.testing.assert_array_equal(log_w, out)


def test_compute_weights_return_prior(proposal, x, log_q):
    """Assert prior is returned"""
    proposal.use_x_prime_prior = False
    log_p = -np.ones(x.size)
    proposal.log_prior = MagicMock(return_value=log_p)
    log_w, log_p_out = BaseFlowProposal.compute_weights(
        proposal, x, log_q, return_log_prior=True
    )

    proposal.log_prior.assert_called_once_with(x)
    expected = -1 - log_q
    np.testing.assert_array_equal(log_w, expected)
    assert log_p_out is log_p


def test_compute_weights_prime_prior(proposal, x, log_q):
    """Test method for computing rejection sampling weights with the prime
    prior.
    """
    proposal.use_x_prime_prior = True
    proposal.x_prime_log_prior = MagicMock(return_value=-np.ones(x.size))
    log_w = BaseFlowProposal.compute_weights(proposal, x, log_q)

    proposal.x_prime_log_prior.assert_called_once_with(x)
    out = -1 - log_q
    np.testing.assert_array_equal(log_w, out)


def test_compute_weights_unit_hypercube(proposal, x, log_q):
    proposal.use_x_prime_prior = False
    proposal.map_to_unit_hypercube = True
    proposal.unit_hypercube_log_prior = MagicMock(
        return_value=-np.ones(x.size)
    )
    log_w = BaseFlowProposal.compute_weights(proposal, x, log_q)

    proposal.unit_hypercube_log_prior.assert_called_once_with(x)
    out = -1 - log_q
    np.testing.assert_array_equal(log_w, out)


def test_compute_acceptance(proposal):
    """Test the compute_acceptance method"""
    proposal.samples = np.arange(1, 11, dtype=float).view([("logL", "f8")])
    acc = BaseFlowProposal.compute_acceptance(proposal, 5.0)
    assert acc == 0.5


def test_convert_to_samples(proposal):
    """Test convert to sample without the prime prior"""
    samples = numpy_array_to_live_points(np.random.randn(10, 2), ["x", "y"])
    proposal.use_x_prime_prior = False
    proposal.model = MagicMock()
    proposal.model.names = ["x"]
    proposal.model.batch_evaluate_log_prior = MagicMock(
        return_value=np.ones(10)
    )

    out_samples = BaseFlowProposal.convert_to_samples(
        proposal, samples, plot=True
    )

    assert out_samples.dtype.names == ("x",) + tuple(
        config.livepoints.non_sampling_parameters
    )


@patch("nessai.proposal.flowproposal.base.plot_1d_comparison")
def test_convert_to_samples_with_prime(mock_plot, proposal):
    """Test convert to sample with the prime prior"""
    samples = numpy_array_to_live_points(np.random.randn(10, 2), ["x", "y"])
    proposal.use_x_prime_prior = True
    proposal.model = MagicMock()
    proposal.model.names = ["x"]
    proposal.model.batch_evaluate_log_prior = MagicMock(
        return_value=np.ones(10)
    )
    proposal._plot_pool = True
    proposal.training_data_prime = "data"
    proposal.output = os.getcwd()
    proposal.populated_count = 1
    proposal.inverse_rescale = MagicMock(return_value=(samples, None))

    out_samples = BaseFlowProposal.convert_to_samples(
        proposal, samples, plot=True
    )

    mock_plot.assert_called_once_with(
        "data",
        samples,
        labels=["live points", "pool"],
        filename=os.path.join(os.getcwd(), "pool_prime_1.png"),
    )
    proposal.inverse_rescale.assert_called_once()
    assert out_samples.dtype.names == ("x",) + tuple(
        config.livepoints.non_sampling_parameters
    )


def test_convert_to_samples_unit_hypercube(proposal):
    """Test convert to sample without the prime prior"""
    samples = numpy_array_to_live_points(np.random.randn(10, 2), ["x", "y"])
    samples_hyper = samples.copy()
    samples["x"] /= 2
    proposal.use_x_prime_prior = False
    proposal.map_to_unit_hypercube = True
    proposal.model = MagicMock()
    proposal.model.names = ["x"]
    proposal.model.batch_evaluate_log_prior = MagicMock(
        return_value=np.ones(10)
    )
    proposal.model.from_unit_hypercube = MagicMock(return_value=samples_hyper)

    out_samples = BaseFlowProposal.convert_to_samples(
        proposal, samples, plot=True
    )

    assert out_samples.dtype.names == ("x",) + tuple(
        config.livepoints.non_sampling_parameters
    )
    np.testing.assert_array_equal(out_samples["x"], samples_hyper["x"])


def test_check_prior_bounds(proposal):
    """Test the check prior bounds method."""
    x = numpy_array_to_live_points(np.arange(10)[:, np.newaxis], ["x"])
    y = np.arange(10)
    proposal.model = Mock()
    proposal.model.in_bounds = MagicMock(
        return_value=np.array(6 * [True] + 4 * [False])
    )
    x_out, y_out = BaseFlowProposal.check_prior_bounds(proposal, x, y)

    assert_structured_arrays_equal(x_out, x[:6])
    np.testing.assert_array_equal(y_out, y[:6])



# -*- coding: utf-8 -*-
"""Test methods related to popluation of the proposal after training"""

import datetime
from functools import partial
from unittest.mock import MagicMock, Mock, call, patch

import numpy as np
import pytest

from nessai.livepoint import (
    get_dtype,
    numpy_array_to_live_points,
    parameters_to_live_point,
)
from nessai.proposal import FlowProposal
from nessai.utils.testing import assert_structured_arrays_equal


@pytest.fixture()
def z(rng):
    return rng.standard_normal((2, 2))


@pytest.fixture()
def x(z, rng):
    return numpy_array_to_live_points(rng.standard_normal(z.shape), ["x", "y"])


@pytest.fixture()
def log_q(x, rng):
    return rng.standard_normal(x.size)


def test_get_alt_distribution_truncated_gaussian(proposal):
    """
    Test getting the alternative distribution for the default latent prior, the
    truncated Gaussian with var=1. This should return None.
    """
    proposal.draw_latent_kwargs = {}
    proposal.latent_prior = "truncated_gaussian"
    dist = FlowProposal.get_alt_distribution(proposal)
    assert dist is None


@pytest.mark.parametrize("prior", ["uniform_nsphere", "uniform_nball"])
def test_get_alt_distribution_uniform(proposal, prior):
    """
    Test getting the alternative distribution for priors that are uniform in
    the n-ball.
    """
    proposal.latent_prior = prior
    proposal.dims = 2
    proposal.r = 2.0
    proposal.fuzz = 1.2
    proposal.flow = Mock()
    proposal.flow.device = "cpu"
    with patch(
        "nessai.proposal.flowproposal.flowproposal.get_uniform_distribution"
    ) as mock:
        dist = FlowProposal.get_alt_distribution(proposal)

    assert dist is not None
    mock.assert_called_once_with(2, 2.4, device="cpu")


def test_radius(proposal):
    """Test computing the radius"""
    z = np.array([[1, 2, 3], [0, 1, 2]])
    expected_r = np.sqrt(14)
    r = FlowProposal.radius(proposal, z)
    np.testing.assert_equal(r, expected_r)


def test_radius_w_log_q(proposal):
    """Test computing the radius with log_q"""
    z = np.array([[1, 2, 3], [0, 1, 2]])
    log_q = np.array([1, 2])
    expected_r = np.sqrt(14)
    r, log_q_r = FlowProposal.radius(proposal, z, log_q)
    assert r == expected_r
    assert log_q_r == log_q[0]


def test_prep_latent_prior_truncated(proposal):
    """Assert prep latent prior calls the correct values"""

    proposal.latent_prior = "truncated_gaussian"
    proposal.dims = 2
    proposal.r = 3.0
    proposal.fuzz = 1.2
    dist = MagicMock()
    dist.sample = MagicMock()

    with patch(
        "nessai.proposal.flowproposal.flowproposal.NDimensionalTruncatedGaussian",  # noqa: E501
        return_value=dist,
    ) as mock_dist:
        FlowProposal.prep_latent_prior(proposal)

    mock_dist.assert_called_once_with(2, 3.0, fuzz=1.2, rng=proposal.rng)

    assert proposal._populate_dist is dist
    assert proposal._draw_func is dist.sample


def test_prep_latent_prior_other(proposal):
    """Assert partial acts as expected"""
    proposal.latent_prior = "gaussian"
    proposal.latent_temperature = 0.9
    proposal.dims = 2
    proposal.r = 3.0
    proposal.fuzz = 1.2

    def draw(dims, N=None, r=None, fuzz=None, rng=None, temperature=None):
        return np.zeros((N, dims))

    proposal._draw_latent_prior = draw

    with patch(
        "nessai.proposal.flowproposal.flowproposal.partial",
        side_effect=partial,
    ) as mock_partial:
        FlowProposal.prep_latent_prior(proposal)

    mock_partial.assert_called_once_with(
        draw, dims=2, r=3.0, fuzz=1.2, rng=proposal.rng, temperature=0.9
    )

    assert proposal._draw_func(N=10).shape == (10, 2)


def test_prep_latent_prior_flow(proposal):
    proposal.latent_prior = "flow"
    proposal.flow = MagicMock()
    proposal.flow.sample_latent_distribution = MagicMock()
    FlowProposal.prep_latent_prior(proposal)
    proposal._draw_func(10)
    proposal.flow.sample_latent_distribution.assert_called_once_with(10)


def test_draw_latent_prior(proposal):
    proposal._draw_func = MagicMock(return_value=[1, 2])
    out = FlowProposal.draw_latent_prior(proposal, 2)
    proposal._draw_func.assert_called_once_with(N=2)
    assert out == [1, 2]


@pytest.mark.parametrize("check_acceptance", [False, True])
@pytest.mark.parametrize("indices", [[], [1]])
@pytest.mark.parametrize("r", [None, 1.0])
@pytest.mark.parametrize(
    "max_radius, min_radius",
    [(0.5, None), (None, 1.5), (None, None)],
)
def test_populate_accumulate_weights(
    proposal,
    check_acceptance,
    indices,
    r,
    min_radius,
    max_radius,
    wait,
    rng,
):
    """Test the main populate method"""
    n_dims = 2
    poolsize = 10
    drawsize = 5
    names = ["x", "y"]
    worst_point = np.array(
        [[1, 2, 3]], dtype=[("x", "f8"), ("y", "f8"), ("logL", "f8")]
    )
    worst_z = rng.standard_normal((1, n_dims))
    z = [
        rng.standard_normal((drawsize, n_dims)),
        rng.standard_normal((drawsize, n_dims)),
        rng.standard_normal((drawsize, n_dims)),
    ]
    x = [
        numpy_array_to_live_points(
            rng.standard_normal((drawsize, n_dims)), names
        ),
        numpy_array_to_live_points(
            rng.standard_normal((drawsize, n_dims)), names
        ),
        numpy_array_to_live_points(
            rng.standard_normal((drawsize, n_dims)), names
        ),
    ]
    log_q = [
        np.log(rng.random(drawsize)),
        np.log(rng.random(drawsize)),
        np.log(rng.random(drawsize)),
    ]
    log_w = [
        np.log(np.concatenate([np.ones(drawsize - 1), np.zeros(1)])),
        np.log(np.concatenate([np.ones(drawsize - 1), np.zeros(1)])),
        np.log(np.concatenate([np.ones(drawsize - 1), np.zeros(1)])),
    ]
    # Control rejection sampling using log_w
    rand_u = 0.5 * np.ones(3 * drawsize)

    log_l = rng.random(poolsize)
    log_p = rng.random(poolsize)

    r_flow = 1.0

    if r is None:
        r_out = r_flow
        if min_radius is not None:
            r_out = max(r_out, min_radius)
        if max_radius is not None:
            r_out = min(r_out, max_radius)
    else:
        r_out = r

    proposal.population_time = datetime.timedelta()
    proposal.initialised = True
    proposal.max_radius = max_radius
    proposal.dims = n_dims
    proposal.poolsize = poolsize
    proposal.drawsize = drawsize
    proposal.min_radius = min_radius
    proposal.fuzz = 1.0
    proposal.indices = indices
    proposal.acceptance = [0.7]
    proposal.keep_samples = False
    proposal.fixed_radius = False
    proposal.compute_radius_with_all = False
    proposal.check_acceptance = check_acceptance
    proposal._plot_pool = True
    proposal.populated_count = 1
    proposal.population_dtype = get_dtype(names)
    proposal.truncate_log_q = False
    proposal.use_x_prime_prior = False
    proposal.accumulate_weights = True

    proposal.forward_pass = MagicMock(return_value=(worst_z, np.nan))
    proposal.backward_pass = MagicMock(side_effect=zip(x, log_q))
    proposal.radius = MagicMock(return_value=r_flow)
    proposal.get_alt_distribution = MagicMock(return_value=None)
    proposal.prep_latent_prior = MagicMock()
    proposal.draw_latent_prior = MagicMock(side_effect=z)
    proposal.latent_prior = "truncated_gaussian"
    proposal.compute_weights = MagicMock(side_effect=log_w)
    proposal.compute_acceptance = MagicMock(return_value=0.8)
    proposal.model = MagicMock()
    proposal.model.batch_evaluate_log_likelihood = MagicMock(
        return_value=log_l
    )
    proposal.rng = MagicMock()
    proposal.rng.random = MagicMock(return_value=rand_u)
    proposal.rng.permutation = rng.permutation

    def convert_to_samples(samples, plot):
        samples["logP"] = log_p
        # wait for windows
        wait()
        return samples

    proposal.plot_pool = MagicMock()
    proposal.convert_to_samples = MagicMock(side_effect=convert_to_samples)

    x_empty = np.empty(0, dtype=proposal.population_dtype)
    with (
        patch(
            "nessai.proposal.flowproposal.flowproposal.empty_structured_array",
            return_value=x_empty,
        ) as mock_empty,
    ):
        FlowProposal.populate(
            proposal, worst_point, n_samples=poolsize, plot=True, r=r
        )

    mock_empty.assert_called_once_with(
        0,
        dtype=proposal.population_dtype,
    )
    proposal.rng.random.assert_called_once_with(3 * drawsize)

    if r is None:
        proposal.forward_pass.assert_called_once_with(
            worst_point,
            rescale=True,
            compute_radius=True,
        )
        proposal.radius.assert_called_once_with(worst_z)
    else:
        assert proposal.r is r

    assert proposal.r == r_out

    proposal.prep_latent_prior.assert_called_once()

    draw_calls = 3 * [call(5)]
    proposal.draw_latent_prior.assert_has_calls(draw_calls)

    backwards_calls = [
        call(
            zz,
            rescale=True,
            return_unit_hypercube=proposal.map_to_unit_hypercube,
        )
        for zz in z
    ]
    proposal.backward_pass.assert_has_calls(backwards_calls)

    compute_weights_calls = [call(xx, lq) for xx, lq in zip(x, log_q)]
    proposal.compute_weights.assert_has_calls(compute_weights_calls)

    proposal.plot_pool.assert_called_once()
    proposal.convert_to_samples.assert_called_once()
    assert_structured_arrays_equal(
        proposal.convert_to_samples.call_args[0][0], proposal.x
    )
    assert proposal.convert_to_samples.call_args[1]["plot"] is True

    assert proposal.population_acceptance == (12 / 15)
    assert proposal.populated_count == 2
    assert proposal.populated is True
    assert proposal.x.size == 10

    if check_acceptance:
        proposal.compute_acceptance.assert_called()
        assert proposal.acceptance == [0.7, 0.8]
    else:
        proposal.compute_acceptance.assert_not_called()

    proposal.model.batch_evaluate_log_likelihood.assert_called_once_with(
        proposal.samples
    )
    np.testing.assert_array_equal(proposal.samples["logL"], log_l)

    assert proposal.population_time.total_seconds() > 0.0


@pytest.mark.parametrize("check_acceptance", [False, True])
@pytest.mark.parametrize("indices", [[], [1]])
@pytest.mark.parametrize("r", [None, 1.0])
@pytest.mark.parametrize(
    "max_radius, min_radius",
    [(0.5, None), (None, 1.5), (None, None)],
)
@pytest.mark.parametrize("latent_prior", ["truncated_gaussian", "gaussian"])
def test_populate_not_accumulate_weights(
    proposal,
    check_acceptance,
    indices,
    r,
    min_radius,
    max_radius,
    latent_prior,
    wait,
    rng,
):
    """Test the main populate method"""
    n_dims = 2
    poolsize = 10
    drawsize = 5
    names = ["x", "y"]
    worst_point = np.array(
        [[1, 2, 3]], dtype=[("x", "f8"), ("y", "f8"), ("logL", "f8")]
    )
    worst_z = rng.standard_normal((1, n_dims))
    z = [
        rng.standard_normal((drawsize, n_dims)),
        rng.standard_normal((drawsize, n_dims)),
        rng.standard_normal((drawsize, n_dims)),
    ]
    x = [
        numpy_array_to_live_points(
            rng.standard_normal((drawsize, n_dims)), names
        ),
        numpy_array_to_live_points(
            rng.standard_normal((drawsize, n_dims)), names
        ),
        numpy_array_to_live_points(
            rng.standard_normal((drawsize, n_dims)), names
        ),
    ]
    log_q = [
        np.log(rng.random(drawsize)),
        np.log(rng.random(drawsize)),
        np.log(rng.random(drawsize)),
    ]
    log_w = [
        np.log(np.concatenate([np.ones(drawsize - 1), np.zeros(1)])),
        np.log(np.concatenate([np.ones(drawsize - 1), np.zeros(1)])),
        np.log(np.concatenate([np.ones(drawsize - 1), np.zeros(1)])),
    ]
    # Control rejection sampling using log_w
    rand_u = [
        0.5 * np.ones(drawsize),
        0.5 * np.ones(drawsize),
        0.5 * np.ones(drawsize),
    ]

    log_l = rng.random(poolsize)
    log_p = rng.random(poolsize)

    r_flow = 1.0

    if r is None:
        r_out = r_flow
        if min_radius is not None:
            r_out = max(r_out, min_radius)
        if max_radius is not None:
            r_out = min(r_out, max_radius)
    else:
        r_out = r

    proposal.population_time = datetime.timedelta()
    proposal.initialised = True
    proposal.max_radius = max_radius
    proposal.dims = n_dims
    proposal.poolsize = poolsize
    proposal.drawsize = drawsize
    proposal.min_radius = min_radius
    proposal.fuzz = 1.0
    proposal.indices = indices
    proposal.acceptance = [0.7]
    proposal.keep_samples = False
    proposal.fixed_radius = False
    proposal.compute_radius_with_all = False
    proposal.check_acceptance = check_acceptance
    proposal._plot_pool = True
    proposal.populated_count = 1
    proposal.population_dtype = get_dtype(names)
    proposal.truncate_log_q = False
    proposal.use_x_prime_prior = False
    proposal.accumulate_weights = False

    proposal.forward_pass = MagicMock(return_value=(worst_z, np.nan))
    proposal.backward_pass = MagicMock(side_effect=zip(x, log_q))
    proposal.radius = MagicMock(return_value=r_flow)
    proposal.get_alt_distribution = MagicMock(return_value=None)
    proposal.latent_prior = latent_prior
    proposal.prep_latent_prior = MagicMock()
    proposal.draw_latent_prior = MagicMock(side_effect=z)
    proposal.compute_weights = MagicMock(side_effect=log_w)
    proposal.compute_acceptance = MagicMock(return_value=0.8)
    proposal.model = MagicMock()
    proposal.model.batch_evaluate_log_likelihood = MagicMock(
        return_value=log_l
    )
    proposal.rng = MagicMock()
    proposal.rng.random = MagicMock(side_effect=rand_u)
    proposal.rng.permutation = rng.permutation

    def convert_to_samples(samples, plot):
        samples["logP"] = log_p
        # wait for windows
        wait()
        return samples

    proposal.plot_pool = MagicMock()
    proposal.convert_to_samples = MagicMock(side_effect=convert_to_samples)

    x_empty = np.empty(poolsize, dtype=proposal.population_dtype)
    with (
        patch(
            "nessai.proposal.flowproposal.flowproposal.empty_structured_array",
            return_value=x_empty,
        ) as mock_empty,
    ):
        FlowProposal.populate(
            proposal, worst_point, n_samples=poolsize, plot=True, r=r
        )

    mock_empty.assert_called_once_with(
        poolsize,
        dtype=proposal.population_dtype,
    )
    proposal.rng.random.assert_has_calls(3 * [call(drawsize)])

    if r is None:
        proposal.forward_pass.assert_called_once_with(
            worst_point,
            rescale=True,
            compute_radius=True,
        )
        proposal.radius.assert_called_once_with(worst_z)
    elif latent_prior == "truncated_gaussian":
        assert proposal.r is r

    if latent_prior == "truncated_gaussian":
        assert proposal.r == r_out
    else:
        assert proposal.r is np.nan

    proposal.prep_latent_prior.assert_called_once()

    draw_calls = 3 * [call(5)]
    proposal.draw_latent_prior.assert_has_calls(draw_calls)

    backwards_calls = [
        call(
            zz,
            rescale=True,
            return_unit_hypercube=proposal.map_to_unit_hypercube,
        )
        for zz in z
    ]
    proposal.backward_pass.assert_has_calls(backwards_calls)

    compute_weights_calls = [call(xx, lq) for xx, lq in zip(x, log_q)]
    proposal.compute_weights.assert_has_calls(compute_weights_calls)

    proposal.plot_pool.assert_called_once()
    proposal.convert_to_samples.assert_called_once()
    assert_structured_arrays_equal(
        proposal.convert_to_samples.call_args[0][0], proposal.x
    )
    assert proposal.convert_to_samples.call_args[1]["plot"] is True

    assert proposal.population_acceptance == (12 / 15)
    assert proposal.populated_count == 2
    assert proposal.populated is True
    assert proposal.x.size == 10

    if check_acceptance:
        proposal.compute_acceptance.assert_called()
        assert proposal.acceptance == [0.7, 0.8]
    else:
        proposal.compute_acceptance.assert_not_called()

    proposal.model.batch_evaluate_log_likelihood.assert_called_once_with(
        proposal.samples
    )
    np.testing.assert_array_equal(proposal.samples["logL"], log_l)

    assert proposal.population_time.total_seconds() > 0.0


def test_populate_not_initialised(proposal):
    """Assert populate fails if the proposal is not initialised"""
    proposal.initialised = False
    with pytest.raises(RuntimeError) as excinfo:
        FlowProposal.populate(proposal, 1.0)
    assert "Proposal has not been initialised. " in str(excinfo.value)


def test_populate_truncate_log_q(proposal, rng):
    n_dims = 2
    nlive = 8
    poolsize = 8
    drawsize = 5
    names = ["x", "y"]
    r_flow = 2.0
    worst_point = np.array(
        [[1, 2, 3]], dtype=[("x", "f8"), ("y", "f8"), ("logL", "f8")]
    )
    z = [
        rng.standard_normal((drawsize, n_dims)),
        rng.standard_normal((drawsize, n_dims)),
        rng.standard_normal((drawsize, n_dims)),
    ]
    x = [
        numpy_array_to_live_points(
            rng.standard_normal((drawsize, n_dims)), names
        ),
        numpy_array_to_live_points(
            rng.standard_normal((drawsize, n_dims)), names
        ),
        numpy_array_to_live_points(
            rng.standard_normal((drawsize, n_dims)), names
        ),
    ]
    log_q = [
        np.zeros(drawsize),
        np.zeros(drawsize),
        np.zeros(drawsize),
    ]
    # This sample will be discarded because of the logq min check
    for i in range(3):
        log_q[i][-1] = np.nan_to_num(-np.inf)
    log_w = [
        np.log(np.concatenate([np.ones(drawsize - 2), np.zeros(1)])),
        np.log(np.concatenate([np.ones(drawsize - 2), np.zeros(1)])),
        np.log(np.concatenate([np.ones(drawsize - 2), np.zeros(1)])),
    ]
    # Control rejection sampling using log_w
    rand_u = 0.5 * np.ones(3 * (drawsize - 1))

    log_l = rng.random(poolsize)

    proposal.population_time = datetime.timedelta()
    proposal.initialised = True
    proposal.dims = n_dims
    proposal.poolsize = poolsize
    proposal.drawsize = drawsize
    proposal.fuzz = 1.0
    proposal.indices = None
    proposal.acceptance = [0.7]
    proposal.keep_samples = False
    proposal.fixed_radius = 2.0
    proposal.compute_radius_with_all = False
    proposal.check_acceptance = False
    proposal._plot_pool = False
    proposal.use_x_prime_prior = False
    proposal.populated_count = 1
    proposal.population_dtype = get_dtype(names)
    proposal.truncate_log_q = True
    proposal.training_data = numpy_array_to_live_points(
        np.random.randn(nlive, n_dims),
        names=names,
    )
    proposal.accumulate_weights = True
    proposal.latent_prior = "truncated_gaussian"

    log_q_live = np.zeros(nlive)
    log_q_live[-1] = -1.0

    proposal.forward_pass = MagicMock(
        return_value=(nlive * [None], log_q_live)
    )
    proposal.backward_pass = MagicMock(side_effect=zip(x, log_q))
    proposal.compute_weights = MagicMock(side_effect=log_w)
    proposal.radius = MagicMock(return_value=r_flow)
    proposal.get_alt_distribution = MagicMock(return_value=None)
    proposal.prep_latent_prior = MagicMock()
    proposal.draw_latent_prior = MagicMock(side_effect=z)
    proposal.rejection_sampling = MagicMock(
        side_effect=[(a[:-1], b[:-1]) for a, b in zip(z, x)]
    )
    proposal.compute_acceptance = MagicMock(return_value=0.8)
    proposal.model = MagicMock()
    proposal.model.batch_evaluate_log_likelihood = MagicMock(
        return_value=log_l
    )
    proposal.rng = MagicMock()
    proposal.rng.random = MagicMock(return_value=rand_u)
    proposal.rng.permutation = rng.permutation

    proposal.convert_to_samples = MagicMock(
        side_effect=lambda *args, **kwargs: args[0]
    )

    x_empty = np.empty(0, dtype=proposal.population_dtype)
    with (
        patch(
            "nessai.proposal.flowproposal.flowproposal.empty_structured_array",
            return_value=x_empty,
        ) as mock_empty,
    ):
        FlowProposal.populate(
            proposal, worst_point, n_samples=poolsize, plot=False
        )

    mock_empty.assert_called_once_with(
        0,
        dtype=proposal.population_dtype,
    )
    proposal.rng.random.assert_called_once_with(3 * drawsize - 3)

    assert proposal.population_acceptance == (9 / 15)

    proposal.forward_pass.assert_called_once_with(proposal.training_data)

    backwards_calls = [
        call(
            zz,
            rescale=True,
            return_unit_hypercube=proposal.map_to_unit_hypercube,
        )
        for zz in z
    ]
    proposal.backward_pass.assert_has_calls(backwards_calls)

    compute_weights_calls = [(xx[:-1], lq[:-1]) for xx, lq in zip(x, log_q)]
    for actual_call, expected_call in zip(
        proposal.compute_weights.call_args_list,
        compute_weights_calls,
    ):
        assert_structured_arrays_equal(actual_call[0][0], expected_call[0])
        np.testing.assert_array_equal(actual_call[0][1], expected_call[1])


def test_populate_enforce_likelihood_threshold(proposal, rng):
    n_dims = 2
    nlive = 8
    drawsize = 5
    names = ["x", "y"]
    r_flow = 2.0
    worst_point = parameters_to_live_point([1, 2], names)
    z = [
        rng.standard_normal((drawsize, n_dims)),
        rng.standard_normal((drawsize, n_dims)),
        rng.standard_normal((drawsize, n_dims)),
    ]
    x = [
        numpy_array_to_live_points(
            rng.standard_normal((drawsize, n_dims)), names
        ),
        numpy_array_to_live_points(
            rng.standard_normal((drawsize, n_dims)), names
        ),
        numpy_array_to_live_points(
            rng.standard_normal((drawsize, n_dims)), names
        ),
    ]

    log_l = [rng.random(drawsize) for _ in range(len(x))]

    log_l_threshold = 0.0
    log_l_accept_total = 0
    while True:
        log_l_accept = [sum(ll > log_l_threshold) for ll in log_l]
        log_l_accept_total = sum(log_l_accept)
        if log_l_accept_total > 5:
            break
        log_l_threshold -= 0.1

    worst_point["logL"] = log_l_threshold
    log_q = [
        np.zeros(drawsize),
        np.zeros(drawsize),
        np.zeros(drawsize),
    ]

    # Reject one from each batch
    log_w = [
        np.log(np.concatenate([np.ones(naccept - 1), np.zeros(1)]))
        for naccept in log_l_accept
    ]
    rand_u = [0.5 * np.ones(naccept) for naccept in log_l_accept]

    poolsize = log_l_accept_total - 3
    proposal.population_time = datetime.timedelta()
    proposal.initialised = True
    proposal.dims = n_dims
    proposal.poolsize = poolsize
    proposal.drawsize = drawsize
    proposal.fuzz = 1.0
    proposal.indices = None
    proposal.acceptance = [0.7]
    proposal.keep_samples = False
    proposal.fixed_radius = None
    proposal.max_radius = None
    proposal.min_radius = None
    proposal.compute_radius_with_all = False
    proposal.check_acceptance = False
    proposal._plot_pool = False
    proposal.use_x_prime_prior = False
    proposal.populated_count = 1
    proposal.population_dtype = get_dtype(names)
    proposal.truncate_log_q = False
    proposal.training_data = numpy_array_to_live_points(
        np.random.randn(nlive, n_dims),
        names=names,
    )
    proposal.accumulate_weights = False
    proposal.latent_prior = "truncated_gaussian"
    proposal.enforce_likelihood_threshold = True
    proposal.log_likelihood_threshold = log_l_threshold

    log_q_live = np.zeros(nlive)

    proposal.forward_pass = MagicMock(
        return_value=(nlive * [None], log_q_live)
    )
    proposal.backward_pass = MagicMock(side_effect=zip(x, log_q))
    proposal.compute_weights = MagicMock(side_effect=log_w)
    proposal.radius = MagicMock(return_value=r_flow)
    proposal.get_alt_distribution = MagicMock(return_value=None)
    proposal.prep_latent_prior = MagicMock()
    proposal.draw_latent_prior = MagicMock(side_effect=z)
    proposal.compute_acceptance = MagicMock(return_value=0.8)
    proposal.model = MagicMock()
    proposal.model.batch_evaluate_log_likelihood = MagicMock(side_effect=log_l)
    proposal.rng = MagicMock()
    proposal.rng.random = MagicMock(side_effect=rand_u)
    proposal.rng.permutation = rng.permutation

    proposal.convert_to_samples = MagicMock(
        side_effect=lambda *args, **kwargs: args[0]
    )

    x_empty = np.empty(poolsize, dtype=proposal.population_dtype)
    with (
        patch(
            "nessai.proposal.flowproposal.flowproposal.empty_structured_array",
            return_value=x_empty,
        ) as mock_empty,
    ):
        FlowProposal.populate(
            proposal, worst_point, n_samples=poolsize, plot=False
        )

    mock_empty.assert_called_once_with(
        poolsize,
        dtype=proposal.population_dtype,
    )
    proposal.rng.random.assert_has_calls(3 * [call(drawsize)])

    assert proposal.population_acceptance == (poolsize / (len(x) * drawsize))

    proposal.forward_pass.assert_called_once_with(
        worst_point, rescale=True, compute_radius=True
    )

    backwards_calls = [
        call(
            zz,
            rescale=True,
            return_unit_hypercube=proposal.map_to_unit_hypercube,
        )
        for zz in z
    ]
    proposal.backward_pass.assert_has_calls(backwards_calls)

    assert proposal.compute_weights.call_count == len(x)
    assert proposal.model.batch_evaluate_log_likelihood.call_count == len(x)



from unittest.mock import MagicMock, patch

import pytest

from nessai.reparameterisations import (
    KnownReparameterisation,
    ReparameterisationDict,
)


def test_known_reparameterisation():
    known = KnownReparameterisation("test", "class", {"key": "value"})
    assert known.name == "test"
    assert known.class_fn == "class"
    assert known.keyword_arguments == {"key": "value"}


def test_known_reparameterisation_no_kwargs():
    known = KnownReparameterisation("test", "class")
    assert known.name == "test"
    assert known.class_fn == "class"
    assert known.keyword_arguments == {}


def test_reparameterisation_dict_add_reparam():
    reparam_dict = ReparameterisationDict()
    reparam_dict.add_reparameterisation("test", "class", {"key": "value"})
    assert reparam_dict["test"].name == "test"
    assert reparam_dict["test"].class_fn == "class"
    assert reparam_dict["test"].keyword_arguments == {"key": "value"}


def test_reparameterisation_dict_add_reparam_existing_entry():
    reparam_dict = ReparameterisationDict(
        {"test": KnownReparameterisation("test", "class", {"key": "value"})}
    )
    with pytest.raises(
        ValueError, match="Reparameterisation test already exists"
    ):
        reparam_dict.add_reparameterisation("test", "class", {"key": "value"})


def test_reparameterisation_dict_add_external_reparam():
    reparam_dict = ReparameterisationDict()
    # Mock class
    external_reparam = KnownReparameterisation(
        "external_reparam", "class", {"key": "value"}
    )

    # Mock what is normally returned by the entry point before they are loaded
    EntryPointClass = MagicMock(spec=["load"])
    EntryPointClass.load = MagicMock(return_value=external_reparam)

    # Always return the version that needs to be loaded
    with patch(
        "nessai.reparameterisations.utils.get_entry_points",
        return_value={"external_class": EntryPointClass},
    ) as mock_get_entry_points:
        reparam_dict.add_external_reparameterisations(
            "nessai.reparameterisations"
        )
    mock_get_entry_points.assert_called_once_with("nessai.reparameterisations")

    assert external_reparam == reparam_dict["external_reparam"]


def test_reparameterisation_dict_add_external_reparam_invalid_type():
    reparam_dict = ReparameterisationDict()
    # Mock class
    external_reparam = ("external_reparam", "class", {"key": "value"})

    # Mock what is normally returned by the entry point before they are loaded
    EntryPointClass = MagicMock(spec=["load"])
    EntryPointClass.load = MagicMock(return_value=external_reparam)

    # Always return the version that needs to be loaded
    with (
        patch(
            "nessai.reparameterisations.utils.get_entry_points",
            return_value={"external_class": EntryPointClass},
        ),
        pytest.raises(
            RuntimeError, match="Invalid external reparameterisation"
        ),
    ):
        reparam_dict.add_external_reparameterisations(
            "nessai.reparameterisations"
        )


def test_reparameterisation_dict_add_external_reparam_name_conflict():
    reparam_dict = ReparameterisationDict()
    reparam_dict.add_reparameterisation("default", "class", {"key": "value"})
    # Mock class
    external_reparam = KnownReparameterisation(
        "default", "class", {"key": "value"}
    )

    # Mock what is normally returned by the entry point before they are loaded
    EntryPointClass = MagicMock(spec=["load"])
    EntryPointClass.load = MagicMock(return_value=external_reparam)

    # Always return the version that needs to be loaded
    with (
        patch(
            "nessai.reparameterisations.utils.get_entry_points",
            return_value={"external_class": EntryPointClass},
        ),
        pytest.raises(
            ValueError, match="Reparameterisation default already exists"
        ),
    ):
        reparam_dict.add_external_reparameterisations(
            "nessai.reparameterisations"
        )



from unittest.mock import MagicMock

import pytest

from nessai.samplers.importancesampler import ImportanceNestedSampler as INS


@pytest.mark.parametrize("met", [False, True])
def test_reached_tolerance(ins, met):
    ins.combined_criterion = MagicMock()
    ins.combined_criterion.is_met.return_value = met
    ins.criterion = {"ess": 1000, "ratio": -0.1}
    assert INS.reached_tolerance.__get__(ins) is met


def test_compute_stopping_criterion(ins):
    ins.combined_criterion = MagicMock()
    ins.combined_criterion.names = ["ratio", "ess"]
    ins.combined_criterion.tolerances = [0.0, 1000]
    ins.state = MagicMock()
    ins.state.ess = 500
    ins.state.ratio = 1.0
    assert INS.compute_stopping_criterion(ins) == {"ratio": 1.0, "ess": 500}


@pytest.mark.parametrize(
    "stopping_criterion, tolerance, check_criteria",
    [
        ("ess", 1000, "all"),
        (["ess", "log_evidence_ratio"], [1000, 0], "all"),
        (["ess", "log_evidence_ratio"], [1000, 0], "any"),
    ],
)
def test_configure_stopping_criterion(
    ins, stopping_criterion, tolerance, check_criteria
):
    INS.configure_stopping_criterion(
        ins, stopping_criterion, tolerance, check_criteria
    )

    if isinstance(stopping_criterion, str):
        stopping_criterion = [stopping_criterion]

    if isinstance(tolerance, (int, float)):
        tolerance = [float(tolerance)]
    else:
        tolerance = [float(t) for t in tolerance]

    assert ins.combined_criterion.names == stopping_criterion
    assert list(ins.combined_criterion.tolerances.values()) == tolerance

    if check_criteria == "all":
        assert ins.combined_criterion.mode == "and"
    elif check_criteria == "any":
        assert ins.combined_criterion.mode == "or"
    else:
        raise ValueError("check_criteria must be 'all' or 'any'")


def test_stopping_criteria(ins):
    ins.combined_criterion = MagicMock()
    ins.combined_criterion.names = ["ratio", "ess"]
    INS.stopping_criteria.__get__(ins) == ins.combined_criterion.names



"""Custom exceptions for the nessai package."""


class RNGError(Exception):
    """Base class for exceptions related to random number generation."""

    pass


class RNGNotSetError(RNGError):
    """Exception raised when the random number generator has not been set"""

    def __init__(
        self, msg="Random number generator has not been set", *args, **kwargs
    ):
        super().__init__(msg, *args, **kwargs)


class RNGSetError(RNGError):
    """Exception raised when the random number generator has already been set"""

    def __init__(
        self,
        msg="Random number generator has already been set",
        *args,
        **kwargs,
    ):
        super().__init__(msg, *args, **kwargs)



"""
Example showing how to handle discrete parameters using reparametersations.
"""

import numpy as np

from nessai.flowsampler import FlowSampler
from nessai.livepoint import empty_structured_array
from nessai.model import Model
from nessai.utils import configure_logger

# Configure the output directory and logger
output = "./outdir/discrete_example/"
logger = configure_logger(output=output, log_level="INFO")

# Set the random number generator
rng = np.random.default_rng(1234)


# Define the model class, this model has two signal models and a discrete
# variable that determines which is used for the computing the log-likelihood
class MultiModelLikelihood(Model):
    def __init__(self, x_data, y_data):
        # x and y data
        self.x_data = x_data
        self.y_data = y_data
        # Assume the standard deviation is known
        self.sigma = 1
        self.bounds = {"m": [-5.0, 5.0], "c": [-5.0, 5.0], "model": [0, 1]}
        self.names = list(self.bounds.keys())
        # Specify the list of discrete parameters
        self.discrete_parameters = ["model"]

    def new_point(self, N=1):
        # Redefine the new point method so that it correctly samples the
        # discrete parameter
        x = empty_structured_array(N, self.names)
        x["m"] = rng.uniform(*self.bounds["m"], size=N)
        x["c"] = rng.uniform(*self.bounds["c"], size=N)
        x["model"] = rng.choice([0, 1], size=N)
        return x

    def new_point_log_prob(self, x):
        # Update the function that defines the log-probability for the
        # `new_point` method.
        log_prob = -np.log(
            np.ptp(self.bounds["m"]) * np.ptp(self.bounds["c"]) * 0.5
        ) * np.ones(len(x))
        return log_prob

    def log_prior(self, x):
        # Check the values are within the prior bounds
        log_p = np.log(self.in_bounds(x), dtype="float")
        # Compute the log-prior for the uniform parameters
        for n in ["m", "c"]:
            log_p -= np.log(self.bounds[n][1] - self.bounds[n][0])
        # Only accept the allowed values for 'model', log(0) = -inf
        log_p += np.log(~(x["model"] % 1).astype(bool))
        return log_p

    def log_likelihood(self, x):
        # Use a different likelihood depending on the value of the "model"
        # parameter. We use the following two models:
        # 1. (mx + c)
        # 2. (mx^1.1 + c)
        y_fit = np.where(
            x["model"] == 1,
            (x["m"] * self.x_data + x["c"]),
            (x["m"] * self.x_data**1.1 + x["c"]),
        )
        # Standard Gaussian likelihood
        log_l = np.sum(
            -0.5 * (((self.y_data - y_fit) / self.sigma) ** 2)
            - np.log(2 * np.pi * self.sigma**2),
            axis=0,
        )
        return log_l


# Generate some fake data.
# We use the straight line model for this example.
x_data = np.linspace(0, 10, 100)
y_data = 2.5 * x_data + 1.4 + rng.standard_normal(len(x_data))

# Define the sampler and specify the 'dequantise' reparameterisation for the
# discrete parameter.
# This adds random noise from [0, 1) to the discrete values, without this
# the sampling would be inefficient since the flow will not propose integer
# values.
fs = FlowSampler(
    MultiModelLikelihood(x_data, y_data),
    output=output,
    resume=False,
    seed=1234,
    reparameterisations={"model": "dequantise"},
)

# And run as usual
fs.run()



"""Reparameterisations for discrete variables"""

import numpy as np

from .rescale import RescaleToBounds


class Dequantise(RescaleToBounds):
    """Reparameterisation that adds noise to discrete variables and then
    rescales to the specified bounds.

    Can also optionally apply a sigmoid/logit transform after rescaling.
    See :py:class:`~nessai.reparameterisations.rescale.RescaleToBounds` for
    more details.

    Note that :code:`update_bounds` is disabled by default and its use is not
    recommended with this reparameterisation.

    Parameters
    ----------
    parameters : list[str]
        List of parameters to apply the reparameterisation to
    prior_bounds : dict
        Dictionary of prior bounds
    rescale_bounds : Optional[list]
        Bounds to rescale to. Defaults to [-1, 1].
    update_bounds : bool
        Update the bounds for rescaling during sampling. Can be enabled but
        not recommended.
    post_rescaling : Optional[str]
        Name of the rescaling to apply after rescaling to the specified bounds.
    """

    def __init__(
        self,
        parameters=None,
        prior_bounds=None,
        rescale_bounds=None,
        update_bounds=False,
        post_rescaling=None,
        rng=None,
    ):
        super().__init__(
            parameters=parameters,
            prior_bounds=prior_bounds,
            prior=None,
            rescale_bounds=rescale_bounds,
            boundary_inversion=None,
            detect_edges=False,
            offset=None,
            update_bounds=update_bounds,
            pre_rescaling=None,
            post_rescaling=post_rescaling,
            rng=rng,
        )

        self.has_pre_rescaling = True
        self.has_prime_prior = False

    def set_bounds(self, prior_bounds):
        super().set_bounds(prior_bounds)
        # The allowed bounds must in in the maximum value + 1
        self.bounds = {
            p: [b[0], b[1] + 1] for p, b in self.prior_bounds.items()
        }

    def pre_rescaling(self, x):
        n = len(x)
        return x + self.rng.random(n), np.zeros(n)

    def pre_rescaling_inv(self, x):
        return np.floor(x), np.zeros(len(x))
